package gate.treetagger;

import java.io.*;
import java.util.*;
import java.net.URL;
import java.lang.*;

import gate.*;
import gate.creole.*;
import gate.util.*;
import gate.util.GateRuntimeException;


/**
 * This class is a wrapper for the language-independent POS tagger from the University of Stuttgart, Germany.
 * It passes GATE a document and a Tree Tagger shell script.
 * Results are stored in the document's TreeTaggerToken annotations
 */
 
 
public class TreeTagger extends AbstractLanguageAnalyser
  implements ProcessingResource {
  
  //the document to be processed
  private gate.Document document;
  
  //the annotation set to be used for the generated annotations
  private String annotationSetName;
  
  // the Tree Tagger shell script
  private URL treeTaggerBinary;

  public Resource init()
    throws ResourceInstantiationException {return this;}

  public void execute() throws ExecutionException {
    //the TreeTagger is only for Linux
    String osName = System.getProperty("os.name").toLowerCase();
    if (osName.indexOf("linux") == -1)
       throw new GateRuntimeException ("The Tree Tagger cannot be run on any other operating systems except Linux.");
    
    if(document == null)
      throw new GateRuntimeException("No document to process!");
    
    if(treeTaggerBinary == null)
      throw new GateRuntimeException("Cannot proceed unless a TreeTagger script is specified.");
   
    //get current text from GATE for the TreeTagger
    File textfile = getCurrentText();
    
    //generate TreeTagger command line
    String treeTaggerCmd = treeTaggerBinary.getFile() + " " + textfile.getAbsolutePath();
    
    //run TreeTagger
    runTreeTagger(treeTaggerCmd );
    
    //delete the temporary text file
    textfile.delete();
  }


  private File getCurrentText(){
     String cText=document.getContent().toString();
     String currentdir = System.getProperty("user.dir");
     File gateTextFile = new File(currentdir+"/treetaggertextfile");
     try{
        FileWriter fw = new FileWriter(gateTextFile);
        fw.write(cText);
        fw.close();   
     } catch (java.io.IOException except){
          System.out.println("File error: "+except);
     }
      return (gateTextFile);
  }
  

  private void runTreeTagger(String cmdline) {
    int startoffset, endoffset;
    String sContent = document.getContent().toString();
    String line, word, tag, comment;
    int indx = 0;
    AnnotationSet aSet;
    
    if (annotationSetName =="") aSet = document.getAnnotations();
    else aSet = document.getAnnotations(annotationSetName);
    
    // run TreeTagger and save output
    try {
      
      Process p = Runtime.getRuntime().exec(cmdline);
                
      //get its tree tagger output (gate input)
      BufferedReader input =
	new BufferedReader
	(new InputStreamReader(p.getInputStream()));
    
      while ((line = input.readLine()) != null) {

	StringTokenizer st = new StringTokenizer(line);
		
	while (st.hasMoreTokens()) {
	  tag = null; comment = null; 
	  word = st.nextToken();
	  if( st.hasMoreTokens() )
	    tag = st.nextToken();
	  if( tag != null )
	    comment = st.nextToken();
	  //tokenFm.put( "Comment", comment );
	  //tokenFm.put( "category" , tag );
	  FeatureMap map = Factory.newFeatureMap();
	  map.put("category", tag);
	  map.put("length", Integer.toString(word.length()));
	  map.put("string", word);
       
          startoffset = sContent.indexOf(word, indx);
       
	  if (startoffset == -1) break;
	  endoffset = startoffset + word.length();
	  indx = endoffset;
	  
	  try {
	    aSet.add(new Long(startoffset), new Long(endoffset), "TreeTaggerToken", map);

	  } catch (InvalidOffsetException ioe) {
	    throw new GateRuntimeException("Invalid offset of the annotation");
	  }
	  
	}
	
      } 
      
    } 
    catch (Exception err) {
      System.out.println( "TreeTagger Exception: " + err );
    }

      
  }


  

  // getter and setter methods
  
  //set the document
  public void setDocument(gate.Document document){
       this.document = document;
  }
  
  //return the document
  public gate.Document getDocument(){
       return document;
  }
 
  //set the annotation set name
  public void setAnnotationSetName(String annotationSetName){
       this.annotationSetName = annotationSetName;
  }
  
  //return the annotation set name
  public String getAnnotationSetName(){
       return annotationSetName;
  }
  
  //set the treeTaggerBinary value
  public void setTreeTaggerBinary(URL treeTaggerBinary) {
    this.treeTaggerBinary = treeTaggerBinary;
  }
 
  //return the treeTaggerBinary value 
  public URL getTreeTaggerBinary() {
    return treeTaggerBinary;
  }
  
} // class TreeTagger
